from django.db import models
from django.core.cache import cache
from django.dispatch import receiver
from django.db.models.signals import post_save
import hashlib
import six


class Placeholder(models.Model):
    key = models.CharField(max_length=40, primary_key=True, db_index=True)
    value = models.TextField(blank=True)

    def __unicode__(self):
        return self.value

    def cache_key(self):
        return "front-edit-%s" % self.key

    @classmethod
    def key_for(cls, name, *bits):
        return hashlib.new('sha1', six.text_type(name + ''.join([six.text_type(token) for token in bits])).encode('utf8')).hexdigest()

    @classmethod
    def copy_content(cls, name, source_bits, target_bits):
        source_key = cls.key_for(name, *source_bits)
        target_key = cls.key_for(name, *target_bits)

        source = cls.objects.filter(key=source_key)
        if source.exists():
            source = source.get()
            cls.objects.create(key=target_key, value=source.value)


class PlaceholderHistory(models.Model):
    placeholder = models.ForeignKey(Placeholder, related_name='history', on_delete=models.CASCADE)
    value = models.TextField(blank=True)
    saved = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('-saved', )

    @property
    def _as_json(self):
        return {'value': self.value, 'saved': self.saved.strftime('%s')}


@receiver(post_save, sender=Placeholder)
def save_placeholder(sender, instance, created, raw, *args, **kwargs):
    if not raw:
        # If we have placeholders, check wheter the content has changed before saving history
        if PlaceholderHistory.objects.filter(placeholder=instance).exists():
            ph = PlaceholderHistory.objects.all()[0]
            if ph.value != instance.value:
                PlaceholderHistory.objects.create(placeholder=instance, value=instance.value)
        else:
            PlaceholderHistory.objects.create(placeholder=instance, value=instance.value)


@receiver(post_save, sender=PlaceholderHistory)
def save_history(sender, instance, created, raw, *args, **kwargs):
    cache.delete(instance.placeholder.cache_key())



import json

from django import template
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.urls import NoReverseMatch, reverse  # NOQA
from django.utils.encoding import smart_str
from django.utils.html import strip_tags

import six
from classytags.arguments import Argument, KeywordArgument, MultiValueArgument
from classytags.core import Options, Tag

from ..conf import settings as django_front_settings
from ..models import Placeholder


register = template.Library()


class FrontEditTag(Tag):
    name = 'front_edit'
    options = Options(
        Argument('name', resolve=False, required=True),
        MultiValueArgument('extra_bits', required=False, resolve=True),
        blocks=[('end_front_edit', 'nodelist')],
    )

    def render_tag(self, context, name, extra_bits, nodelist=None):
        hash_val = Placeholder.key_for(name, *extra_bits)
        cache_key = "front-edit-%s" % hash_val

        val = cache.get(cache_key)
        if val is None:
            try:
                val = Placeholder.objects.get(key=hash_val).value
                cache.set(cache_key, val, 3600 * 24)
            except Placeholder.DoesNotExist:
                pass

        if val is None and nodelist:
            val = nodelist.render(context)

        classes = ['editable']

        if getattr(django_front_settings, 'DJANGO_FRONT_EXTRA_CONTAINER_CLASSES', None):
            classes.append(
                six.text_type(django_front_settings.DJANGO_FRONT_EXTRA_CONTAINER_CLASSES)
            )

        user = context.get('request', None) and context.get('request').user
        if django_front_settings.DJANGO_FRONT_PERMISSION(user):
            render = six.text_type(smart_str(val)).strip()
            if not strip_tags(render).strip():
                classes.append('empty-editable')

            return '<div class="%s" id="%s">%s</div>' % (
                ' '.join(classes),
                hash_val,
                render,
            )
        return val or ''


class FrontEditJS(Tag):
    name = 'front_edit_scripts'
    options = Options(
        KeywordArgument('editor', resolve=False, required=False, defaultkey='editor')
    )

    def render_tag(self, context, editor=''):
        try:
            save_url = reverse('front-placeholder-save')
            history_url = reverse('front-placeholder-history', args=('0000',))
        except NoReverseMatch:
            raise ImproperlyConfigured(
                'You must add an urlconf entry for django-front to work, see: http://django-front.readthedocs.org/en/latest/installation.html'
            )

        static_url = context.get('STATIC_URL', '/static/')
        user = context.get('request', None) and context.get('request').user
        token = six.text_type(context.get('csrf_token'))
        plugin = (
            editor.get('editor').lower()
            if editor.get('editor')
            and editor.get('editor').lower()
            in django_front_settings.DJANGO_FRONT_ALLOWED_EDITORS
            else 'default'
        )
        edit_mode = (
            django_front_settings.DJANGO_FRONT_EDIT_MODE
            if django_front_settings.DJANGO_FRONT_EDIT_MODE in ('lightbox', 'inline')
            else 'lightbox'
        )

        if django_front_settings.DJANGO_FRONT_PERMISSION(user):
            return """
<link rel="stylesheet" href="%sfront/css/front-edit.css" />
<script>
    document._front_edit = {
        save_url: '%s',
        history_url_prefix: '%s',
        csrf_token: '%s',
        plugin: '%s',
        static_root: '%s',
        edit_mode: '%s',
        editor_options: %s
    };
</script>
<script src="%sfront/js/front-edit.%s.js"></script>
<script src="%sfront/js/front-edit.js"></script>""".strip() % (
                static_url,
                save_url,
                history_url.replace('0000/', ''),
                token,
                plugin,
                static_url,
                edit_mode,
                json.dumps(django_front_settings.DJANGO_FRONT_EDITOR_OPTIONS),
                static_url,
                plugin,
                static_url,
            )
        else:
            return ''


register.tag(FrontEditTag)
register.tag(FrontEditJS)



from django.http import HttpResponse
try:
    from django.template import engines
    __is_18 = True
except ImportError:
    from django.template import loader, RequestContext
    __is_18 = False


TEST_TEMPLATE = r'''
{% load front_tags %}
<!DOCTYPE html>
<html lang="{{LANGUAGE_CODE}}">
    <head>
        <meta http-equiv="Content-type" content="text/html; charset=utf-8">
        <title>front test</title>
    </head>
    <body>
        <p>Hello, {% if request.user.is_authenticated %}{{request.user}}{% else %}Anon{% endif %}!</p>
        <div id="global-ph">{% front_edit "global-ph" %}global base content{% end_front_edit  %}</div>
        <div id="locale-ph">{% front_edit "locale-ph" request.LANGUAGE_CODE %}locale base content{% end_front_edit %}</div>
        <div id="some-other-ph">{% front_edit "some-other-ph" arg1 %}argument based content{% end_front_edit %}</div>
        {% front_edit_scripts editor="ace" %}
    </body>
</html>
'''


TEST_TEMPLATE_INVALID_EDITOR = r'''
{% load front_tags %}
{% front_edit_scripts editor="dummy" %}
'''


def _render(template_string, context, request):
    if __is_18:
        return engines['django'].from_string(template_string).render(context=context, request=request)
    else:
        return loader.get_template_from_string(template_string).render(RequestContext(request, context))


def test(request):
    return HttpResponse(_render(TEST_TEMPLATE, context=dict(arg1='hello'), request=request))


def test_invalid_template_tag(request):
    return HttpResponse(_render(TEST_TEMPLATE_INVALID_EDITOR, context={}, request=request))



# Generated by Django 2.0a1 on 2017-09-23 14:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('front', '0002_auto_20150123_0219'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='placeholderhistory',
            options={'ordering': ('-saved',)},
        ),
    ]



from .conf import settings as django_front_settings
from .models import Placeholder, PlaceholderHistory
from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
from django.views.decorators.http import require_POST, require_GET
import json


class JsonHttpResponse(HttpResponse):
    def __init__(self, **kwargs):
        super(JsonHttpResponse, self).__init__(json.dumps(kwargs, cls=DjangoJSONEncoder, ensure_ascii=False), content_type='application/json')


@require_POST
def do_save(request):
    if request.POST and django_front_settings.DJANGO_FRONT_PERMISSION(request.user):
        key, val = request.POST.get('key'), request.POST.get('val')
        placeholder, created = Placeholder.objects.get_or_create(key=key, defaults=dict(value=val))
        if not created:
            placeholder.value = val
            placeholder.save()

        return HttpResponse('1')
    return HttpResponse('0')


@require_GET
def get_history(request, key):
    return JsonHttpResponse(history=[ph._as_json for ph in PlaceholderHistory.objects.filter(placeholder__key=key)])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('front', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='placeholderhistory',
            name='placeholder',
            field=models.ForeignKey(related_name='history', to='front.Placeholder', on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



import django

from .views import do_save, get_history


if django.VERSION >= (3, 1, 0):
    from django.urls import re_path as url
else:
    from django.conf.urls import url


urlpatterns = [
    url(r'^save/$', do_save, name='front-placeholder-save'),
    url(
        r'^hist/(?P<key>[0-9a-f]{1,40})/$', get_history, name='front-placeholder-history'
    ),
]



# -*- coding: utf-8 -*-
import json
import re

import six
from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from django.urls import reverse
from front.models import Placeholder, PlaceholderHistory
from front.conf import settings as django_front_settings


@override_settings(ROOT_URLCONF='front.tests.urls')
class FrontTestCase(TestCase):

    def setUp(self):
        self.admin_user = User.objects.create_user('admin_user', 'admin@admin.com', 'admin_user')
        self.admin_user.is_staff = True
        self.admin_user.is_superuser = True
        self.admin_user.save()

    def tearDown(self):
        cache.clear()

    def test_1_anonymous_user_cant_see_tags(self):
        resp = self.client.get(reverse('front-test'))
        self.assertTrue('document._front_edit' not in six.text_type(resp.content))

    def test_2_root_user_does(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertTrue('document._front_edit' in six.text_type(resp.content))
        self.assertTrue(six.text_type("plugin: 'ace'") in six.text_type(resp.content.decode('utf8')))

    def test_3_anonymous_user_cant_post_either(self):
        resp = self.client.post(reverse('front-placeholder-save'), {'key': '123123', 'val': '<p>booh!</p>'})
        self.assertTrue('0' in six.text_type(resp.content))
        self.assertEqual(0, Placeholder.objects.count())

    def test_4_but_admin_shall_post(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.post(reverse('front-placeholder-save'), {'key': '123123', 'val': '<p>booh!</p>'})
        self.assertTrue('1' in six.text_type(resp.content))
        self.assertEqual(1, Placeholder.objects.count())

    def test_5_workflow_baby(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        ids = re.findall(r'<div class="editable" id="([^"]+)">', six.text_type(resp.content))
        self.assertTrue(len(ids))

        self.assertTrue('global base content' in six.text_type(resp.content))
        resp = self.client.post(reverse('front-placeholder-save'), {'key': ids[0], 'val': '<p>booh!</p>'})
        resp = self.client.post(reverse('front-placeholder-save'), {'key': ids[1], 'val': '<p>english!</p>'})

        resp = self.client.get(reverse('front-test'))
        self.assertFalse('global base content' in six.text_type(resp.content))
        self.assertTrue('booh!' in six.text_type(resp.content))
        self.assertTrue('english!' in six.text_type(resp.content))
        self.assertTrue('locale base content' not in six.text_type(resp.content))

        # Parlez vous francais?
        # The global placholder should display the same in english, but not the localized one
        self.client.post('/i18n/setlang/', {'language': 'fr'})
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('global base content' in six.text_type(resp.content))
        self.assertTrue('booh!' in six.text_type(resp.content))
        self.assertFalse('english!' in six.text_type(resp.content))
        self.assertFalse('locale base content' not in six.text_type(resp.content))

    def test_6_warn_when_missing_urlconf(self):
        with self.settings(ROOT_URLCONF='front.tests.urls_no_save'):
            self.client.login(username='admin_user', password='admin_user')
            try:
                self.client.get(reverse('front-test'))
                self.fail()
            except ImproperlyConfigured:
                pass

    def test_7_empty_content(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('empty-editable' in six.text_type(resp.content))
        ids = re.findall(r'<div class="editable" id="([^"]+)">', six.text_type(resp.content))

        resp = self.client.post(reverse('front-placeholder-save'), {'key': ids[0], 'val': ''})
        self.assertTrue('1' in six.text_type(resp.content))
        resp = self.client.get(reverse('front-test'))
        self.assertTrue('empty-editable' in six.text_type(resp.content))

        resp = self.client.post(reverse('front-placeholder-save'), {'key': ids[0], 'val': '<p>    </p>'})
        self.assertTrue('1' in six.text_type(resp.content))
        resp = self.client.get(reverse('front-test'))
        self.assertTrue('empty-editable' in six.text_type(resp.content))

        resp = self.client.post(reverse('front-placeholder-save'), {'key': ids[0], 'val': '<p>booh</p>'})
        self.assertTrue('1' in six.text_type(resp.content))
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('empty-editable' in six.text_type(resp.content))

    def test_8_save_history(self):
        self.assertEqual(0, PlaceholderHistory.objects.count())

        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('empty-editable' in six.text_type(resp.content))
        ids = re.findall(r'<div class="editable" id="([^"]+)">', six.text_type(resp.content))
        key = ids[0]

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh</p>'})
        self.assertTrue('1' in six.text_type(resp.content))
        resp = self.client.get(reverse('front-test'))

        self.assertEqual(1, PlaceholderHistory.objects.filter(placeholder__key=key).count())

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh, too</p>'})
        resp = self.client.get(reverse('front-test'))

        self.assertEqual(2, PlaceholderHistory.objects.filter(placeholder__key=key).count())

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': ''})
        self.assertEqual(3, PlaceholderHistory.objects.filter(placeholder__key=key).count())

    def test_9_get_history(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('empty-editable' in six.text_type(resp.content))
        ids = re.findall(r'<div class="editable" id="([^"]+)">', six.text_type(resp.content))
        key = ids[0]

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh</p>'})
        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh, too</p>'})
        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': ''})
        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': u'<p>aéaéaàà</p>'})

        self.assertEqual(4, PlaceholderHistory.objects.filter(placeholder__key=key).count())

        resp = self.client.get(reverse('front-placeholder-history', args=(key, )))
        self.assertTrue('application/json' in resp['Content-Type'])

        self.assertTrue(u'<p>aéaéaàà</p>' in [ph.get('value') for ph in json.loads(resp.content.decode("utf8")).get('history')])

    def test_10_history_only_saves_changes(self):

        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('empty-editable' in six.text_type(resp.content))
        ids = re.findall(r'<div class="editable" id="([^"]+)">', six.text_type(resp.content))
        key = ids[0]

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh</p>'})
        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh, too</p>'})

        self.assertEqual(2, PlaceholderHistory.objects.filter(placeholder__key=key).count())

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh, too</p>'})
        self.assertEqual(2, PlaceholderHistory.objects.filter(placeholder__key=key).count())

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh</p>'})
        self.assertEqual(3, PlaceholderHistory.objects.filter(placeholder__key=key).count())

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh</p>'})
        self.assertEqual(3, PlaceholderHistory.objects.filter(placeholder__key=key).count())

    def test_11_invalid_tag(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test_invalid_template_tag'))
        self.assertTrue(six.text_type(u"plugin: 'default'") in six.text_type(resp.content.decode('utf8')))

    def test_12_calculate_keys(self):
        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertFalse('empty-editable' in six.text_type(resp.content))
        ids = re.findall(r'<div class="editable" id="([^"]+)">', six.text_type(resp.content))
        self.assertEqual(six.text_type(ids[0]), six.text_type('d577f15230caa8d39fb651d5b1ea34743f56edff'))

    def test_13_copy_content(self):
        self.assertEqual(0, Placeholder.objects.count())
        self.client.login(username='admin_user', password='admin_user')

        resp = self.client.get(reverse('front-test'))
        key = Placeholder.key_for('global-ph')
        self.assertTrue(key in six.text_type(resp.content))
        key = Placeholder.key_for('some-other-ph', 'hello')
        self.assertTrue(key in six.text_type(resp.content))

        resp = self.client.post(reverse('front-placeholder-save'), {'key': key, 'val': '<p>booh</p>'})
        self.assertEqual(1, Placeholder.objects.count())

        Placeholder.copy_content('some-other-ph', ['hello'], ['jello'])
        self.assertEqual(2, Placeholder.objects.count())

        jello_key = Placeholder.key_for('some-other-ph', 'jello')
        hello_key = Placeholder.key_for('some-other-ph', 'hello')
        self.assertTrue(Placeholder.objects.filter(key=jello_key).exists())
        self.assertTrue(Placeholder.objects.filter(key=hello_key).exists())
        self.assertEqual(
            Placeholder.objects.get(key=jello_key).value,
            Placeholder.objects.get(key=jello_key).value
        )

    def test_14_extra_container_classes(self):
        _setting = getattr(django_front_settings, 'DJANGO_FRONT_EXTRA_CONTAINER_CLASSES', '')
        django_front_settings.DJANGO_FRONT_EXTRA_CONTAINER_CLASSES = 'some extra classes'

        self.client.login(username='admin_user', password='admin_user')
        resp = self.client.get(reverse('front-test'))
        self.assertTrue('some extra classes' in six.text_type(resp.content))

        django_front_settings.DJANGO_FRONT_EXTRA_CONTAINER_CLASSES = _setting



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import render_to_response
from django.http import HttpResponse, HttpResponseRedirect
from django.template import RequestContext


def home(request):
    return render_to_response('base.html', dict(
    ), context_instance=RequestContext(request))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Placeholder',
            fields=[
                ('key', models.CharField(max_length=40, serialize=False, primary_key=True, db_index=True)),
                ('value', models.TextField(blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PlaceholderHistory',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('value', models.TextField(blank=True)),
                ('saved', models.DateTimeField(auto_now_add=True)),
                ('placeholder', models.ForeignKey(to='front.Placeholder', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': (b'-saved',),
            },
            bases=(models.Model,),
        ),
    ]



import django

from .views import test


if django.VERSION >= (3, 1, 0):
    from django.urls import re_path as url, include
else:
    from django.conf.urls import url, include

urlpatterns = [
    url(r'^test/$', test, name='front-test'),
    url(r'^i18n/', include('django.conf.urls.i18n')),
]



from django.conf import settings

DJANGO_FRONT_PERMISSION = getattr(settings, 'DJANGO_FRONT_PERMISSION', lambda u: u and u.is_staff)
DJANGO_FRONT_EDIT_MODE = getattr(settings, 'DJANGO_FRONT_EDIT_MODE', 'lightbox')  # ('lightbox' or 'inline')
DJANGO_FRONT_EDITOR_OPTIONS = getattr(settings, 'DJANGO_FRONT_EDITOR_OPTIONS', dict())
DJANGO_FRONT_ALLOWED_EDITORS = [editor.lower() for editor in getattr(settings, 'DJANGO_FRONT_ALLOWED_EDITORS', ['ace', 'ace-local', 'wymeditor', 'redactor', 'epiceditor', 'ckeditor', 'default', 'froala', 'medium', 'summernote'])]
DJANGO_FRONT_EXTRA_CONTAINER_CLASSES = getattr(settings, 'DJANGO_FRONT_EXTRA_CONTAINER_CLASSES', '')



# Django settings for test_project project.
import os

DEBUG = True
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(os.path.dirname(__file__), "test_project.db"),
    }
}

USE_I18N = True
SECRET_KEY = "lol I dont even care"
USE_TZ = True

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.locale.LocaleMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": (
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ),
            "debug": False,
        },
    }
]
STATIC_URL = "/static/"

ROOT_URLCONF = "test_project.urls"

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = "test_project.wsgi.application"

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "front",
)

SITE_ID = 1



import django

from .views import test, test_invalid_template_tag


if django.VERSION >= (3, 1, 0):
    from django.urls import re_path as url, include
else:
    from django.conf.urls import url, include

urlpatterns = [
    url(r'^test/$', test, name='front-test'),
    url(
        r'^test-invalid/$',
        test_invalid_template_tag,
        name='front-test_invalid_template_tag',
    ),
    url(r'^i18n/', include('django.conf.urls.i18n')),
    url(r'^front-edit/', include('front.urls')),
]
